# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from .lvqt_real import LVQT as _LVQT
from .utils import *

# Regular imports
import torch


class LVQT(_LVQT):
    """
    Implements an extension of the real-only LVQT variant. In this version, the imaginary
    weights of the transform are inferred from the real weights by using the Hilbert transform.
    """

    def __init__(self, **kwargs):
        """
        Initialize LVQT parameters and the PyTorch processing modules.

        Parameters
        ----------
        See _LVQT class...
        """

        super(LVQT, self).__init__(**kwargs)

        # Initialize l2 pooling to recombine real/imag filter channels
        self.l2_pool = torch.nn.LPPool1d(norm_type=2, kernel_size=2, stride=2)

    def forward(self, audio):
        """
        Perform the main processing steps for the filterbank.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        feats : Tensor (B x F x T)
          Features calculated for a batch of tracks,
          B - batch size
          F - dimensionality of features (number of bins)
          T - number of time steps (frames)
        """

        # Run the standard convolution steps (call grandparent function)
        real_feats = super(type(self).__bases__[0], self).forward(audio)

        # We manually do the padding for the convolutional
        # layer to allow for different front/back padding
        padded_audio = torch.nn.functional.pad(audio, self.pd1)

        # Obtain the imaginary weights
        imag_weights = self.get_imag_weights().unsqueeze(1)
        # Convolve the audio with the filterbank of imaginary weights
        imag_feats = torch.nn.functional.conv1d(padded_audio,
                                                weight=imag_weights,
                                                stride=self.sd1,
                                                padding=0)

        # Add an extra dimension to both sets of features
        real_feats = real_feats.unsqueeze(-1)
        imag_feats = imag_feats.unsqueeze(-1)

        # Concatenate the features along a new dimension
        feats = torch.cat((real_feats, imag_feats), dim=-1)
        # Switch filter and frame dimension
        feats = feats.transpose(1, 2)
        # Collapse the last dimension to zip the features
        # and make the real/imag responses adjacent
        feats = feats.reshape(tuple(list(feats.shape[:-2]) + [-1]))
        # Perform l2 pooling across the filter dimension
        feats = self.l2_pool(feats)
        # Switch the frame and filter dimension back
        feats = feats.transpose(1, 2)

        # Perform post-processing steps
        feats = self.post_proc(feats)

        return feats

    def get_imag_weights(self):
        """
        Obtain the weights of the imaginary part of the transform using Hilbert transform.

        Returns
        ----------
        imag_weights : Tensor (F x T)
          Weights of the imaginary part of the transform,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        real_weights = self.get_real_weights()

        if self.update:
            # Calculate HT as normal
            imag_weights = torch_hilbert(real_weights)
        else:
            # Don't allow gradients to propagate through HT
            with torch.no_grad():
                imag_weights = torch_hilbert(real_weights)

        return imag_weights



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# Regular imports
import numpy as np
import librosa
import torch

EPSILON = torch.finfo(torch.float32).eps


def torch_amplitude_to_db(feats, amin=1e-10, top_db=80.0, to_prob=False):
    """
    Take the log of a time-frequency representation with differentiable
    PyTorch functions, scaling values to be between 0 and 1. Adapted from
    the Librosa amplitude_to_db function.

    Parameters
    ----------
    feats : Tensor (B x H x F x T) or (B x F x T)
      Set of amplitude features,
      B - batch size
      H - number of harmonics (a.k.a. channels)
      F - dimensionality of features
      T - number of time steps (frames)
    amin : float
      Minimum amplitude possible within a time-frequency bin
    top_db : float
      Maximum difference from dB ceiling (defines dB floor)
    to_prob : bool
      Scale decibel values to be between 0 and 1

    Returns
    ----------
    log_feats : Tensor (B x H x F x T) or (B x F x T)
      Set of dB or probability features
    """

    # Perform processing on a copy of original features
    feats_copy = feats.clone()

    # Get handles for the feature dimensionality
    B, H, F, T = feats.size(0), 1, feats.size(-2), feats.size(-1)

    # Collapse the channel dimension if it exists
    if len(feats.size()) > 3:
        H = feats.size(1)
        feats_copy = feats_copy.view(B * H, F, T)

    # Get handle for pseudo batch size
    PB = B * H

    # Convert amplitude features to power
    power = torch.abs(feats_copy) ** 2

    # Get reference values (maximums) from power spectrogram for each 2D transform
    ref_value = torch.max(power.view(PB, -1), dim=-1)[0]

    # Clamp power spectrogram at specified minimum - effectively max(amin, power)
    power[power < amin] = amin

    # Convert power to dB
    log_feats = 10.0 * torch.log10(power)

    # Make sure reference values are above minimum amplitude
    amin = torch.Tensor([amin] * PB).to(power.device)
    amin[amin < ref_value] = ref_value[amin < ref_value]
    # Add dimensions for broadcasting
    amin = amin.unsqueeze(-1).unsqueeze(-1)

    # Combined with previous log, we are performing 10 * log10(power / ref)
    log_feats = log_feats - 10.0 * torch.log10(amin)

    # Collapse the last two dimensions temporarily to make the following easier
    log_feats = log_feats.view(PB, -1)
    # Determine the dB floor for each 2D transform
    dB_floor = (torch.max(log_feats, dim=-1)[0] - top_db).unsqueeze(-1)
    # Clamp the values at the specified dB floor
    log_feats[log_feats < dB_floor] = dB_floor.repeat(1, F * T)[log_feats < dB_floor]

    if to_prob:
        # Scale values and offset to be between 0 and 1
        log_feats = (log_feats / top_db) + 1.0

    # Reshape the log-scaled features using the original feature dimensions
    log_feats = log_feats.view(feats.size())

    return log_feats


def torch_hilbert(x_real, n_fft=None):
    """
    Obtain imaginary counterpart to a real signal such that there are no negative frequency
    components when represented as a complex signal. This is done by using the Hilbert transform.
    We end up with an analytic signal and return only the imaginary part. Most importantly,
    this procedure is fully differentiable. Adapted from the SciPy signal.hilbert function.

    Parameters
    ----------
    x_real : Tensor (F x T)
      Real counterpart of an analytic signal,
      F - number of independent signals
      T - number of time steps (samples)
    n_fft : int
      Number of Fourier components

    Returns
    ----------
    x_imag : Tensor (F x T)
      Imaginary counterpart of an analytic signal,
      F - number of independent signals
      T - number of time steps (samples)
    """

    # Default to the length of the input signal
    if n_fft is None:
        n_fft = x_real.size(-1)

    # Create the transfer function for an analytic signal
    h = torch.zeros(n_fft).to(x_real.device)
    if n_fft % 2 == 0:
        h[0] = h[n_fft // 2] = 1
        h[1 : n_fft // 2] = 2
    else:
        h[0] = 1
        h[1 : (n_fft + 1) // 2] = 2

    # Take the Fourier transform of the real part
    Xf = torch.fft.fft(x_real, n=n_fft, dim=-1)
    # Apply the transfer function to the Fourier transform
    Xfh = Xf * h.unsqueeze(-2)
    # Take the inverse Fourier Transform to obtain the analytic signal
    x_alyt = torch.fft.ifft(Xfh, dim=-1)
    # Take the imaginary part of the analytic signal to obtain the Hilbert transform
    x_imag = x_alyt.imag

    return x_imag


def fft_response(signal, sample_rate, n_fft=None, decibels=False):
    """
    Obtain a signal's FFT response in a plot-friendly format.

    Parameters
    ----------
    signal : ndarray
      Signal to transform
    sample_rate : int
      Number of samples per second
    n_fft : int or None (optional)
      See np.fft.fft documentation...
    decibels : bool
      Whether to convert to dB or leave as amplitude

    Returns
    ----------
    frequencies : ndarray
      Ordered frequencies corresponding to each response
    response : ndarray
      Magnitude response of signal for each basis
    """

    # Take the FFT and calculate the magnitude of the response
    response = np.abs(np.fft.fft(signal, n=n_fft))

    if decibels:
        # Convert the frequency response from amplitude to decibels
        response = librosa.amplitude_to_db(response, ref=np.max)

    # Determine the number of bins in the response
    num_bins = response.shape[-1]

    # Get the frequencies corresponding to the FFT indices
    frequencies = np.fft.fftfreq(num_bins, (1 / sample_rate))

    # Re-order the FFT and frequencies so they go from most negative to most positive
    response = np.roll(response, response.shape[-1] // 2, axis=-1)
    frequencies = np.roll(frequencies, frequencies.shape[-1] // 2)

    return frequencies, response


def filter_kwargs(keywords, **kwargs):
    """
    Filter provided keyword arguments and remove all but
    those matching a provided list of relevant keywords.

    Parameters
    ----------
    keywords : list of string
      Keywords to search for within the provided keyword arguments
    kwargs : dict
      Dictionary of provided keywords

    Returns
    ----------
    filtered_kwargs : dict
      Keyword arguments matching provided keyword list entries
    """

    # Create an empty dictionary for keyword matches
    filtered_kwargs = dict()

    # Loop through relevant keywords
    for key in keywords:
        # Check if there is a match
        if key in kwargs:
            # Add the keyword argument to the matches dictionary
            filtered_kwargs[key] = kwargs[key]

    return filtered_kwargs


def tensor_to_array(tensor):
    """
    Simple helper function to convert a PyTorch tensor
    into a NumPy array in order to keep code readable.

    Parameters
    ----------
    tensor : PyTorch tensor
      Tensor to convert to array

    Returns
    ----------
    array : NumPy ndarray
      Converted array
    """

    # Change device to CPU,
    # detach from gradient graph,
    # and convert to NumPy array
    array = tensor.cpu().detach().numpy()

    return array


def array_to_tensor(array, device=None):
    """
    Simple helper function to convert a NumPy array
    into a PyTorch tensor in order to keep code readable.

    Parameters
    ----------
    array : NumPy ndarray
      Array to convert to tensor
    device : string, or None (optional)
      Add tensor to this device, if specified

    Returns
    ----------
    tensor : PyTorch tensor
      Converted tensor
    """

    # Convert to PyTorch tensor
    tensor = torch.from_numpy(array)

    # Add tensor to device, if specified
    if device is not None:
        tensor = tensor.to(device)

    return tensor



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from .lvqt import _LVQT
from .variational import *

# Regular imports
import numpy as np
import torch


class LVQT(_LVQT):
    """
    Implements a slight adaptation/modernization of the original module presented in my Master's
    Thesis (https://scholarworks.rit.edu/theses/10143/). This variant is referred to as the
    classic variant.
    """

    def __init__(self, **kwargs):
        """
        Initialize LVQT parameters and the PyTorch processing modules.

        Parameters
        ----------
        See _LVQT class...
        """

        super(LVQT, self).__init__(**kwargs)

        # One channel (audio) coming in
        nf_in = 1
        # Two channels (real/imag) for each bin going out
        nf_out = 2 * self.n_bins

        # Initialize the 1D convolutional filterbank
        if self.var_drop:
            self.time_conv = VariationalDropoutConv1d(in_channels=nf_in, out_channels=nf_out,
                                                      kernel_size=self.ks1, stride=self.sd1,
                                                      log_sigma2=self.var_drop)
        else:
            self.time_conv = torch.nn.Conv1d(in_channels=nf_in, out_channels=nf_out,
                                             kernel_size=self.ks1, stride=self.sd1,
                                             bias=False)

        if not self.random:
            # Split the complex valued bases into real and imaginary weights
            real_weights, imag_weights = np.real(self.basis), np.imag(self.basis)
            # Zip them together as separate channels so both components of each filter are adjacent
            complex_weights = np.array([[real_weights[i]] + [imag_weights[i]]
                                        for i in range(self.n_bins)])
            # View the real/imag channels as independent filters
            complex_weights = torch.Tensor(complex_weights).view(nf_out, 1, self.ks1)
            # Manually set the Conv1d parameters with the real/imag weights
            self.time_conv.weight = torch.nn.Parameter(complex_weights)

        # Initialize l2 pooling to recombine real/imag filter channels
        self.l2_pool = torch.nn.LPPool1d(norm_type=2, kernel_size=2, stride=2)

    def forward(self, audio):
        """
        Perform the main processing steps for the filterbank.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        feats : Tensor (B x F x T)
          Features calculated for a batch of tracks,
          B - batch size
          F - dimensionality of features (number of bins)
          T - number of time steps (frames)
        """

        # Run the standard convolution steps
        feats = super().forward(audio)

        # Switch filter and frame dimension
        feats = feats.transpose(1, 2)
        # Perform l2 pooling across the filter dimension
        feats = self.l2_pool(feats)
        # Switch the frame and filter dimension back
        feats = feats.transpose(1, 2)

        # Perform post-processing steps
        feats = self.post_proc(feats)

        return feats

    def get_weights(self):
        """
        Obtain the weights of the transform split by real/imag.

        Returns
        ----------
        comp_weights : Tensor (F x 2 x T)
          Weights of the transform split by real/imag,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        weights = self.time_conv.weight
        weights = weights.view(self.n_bins, 2, -1)

        return weights

    def get_real_weights(self):
        """
        Obtain the weights of the real part of the transform.

        Returns
        ----------
        real_weights : Tensor (F x T)
          Weights of the real part of the transform,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        comp_weights = self.get_weights()
        real_weights = comp_weights[:, 0]

        return real_weights

    def get_imag_weights(self):
        """
        Obtain the weights of the imaginary part of the transform.

        Returns
        ----------
        imag_weights : Tensor (F x T)
          Weights of the imaginary part of the transform,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        comp_weights = self.get_weights()
        imag_weights = comp_weights[:, 1]

        return imag_weights



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from .utils import *

# Regular imports
from mpl_toolkits.axisartist.axislines import SubplotZero
from librosa.filters import constant_q
from matplotlib import pyplot as plt
from matplotlib import rcParams
from abc import abstractmethod

import soundfile as sf
import numpy as np
import warnings
import librosa
import torch
import math
import os

# TODO - savefig() makes plotting slower than it could be. Can this be optimized?


class _LVQT(torch.nn.Module):
    """
    Abstract class to implement common functionality across LVQT variants.
    """

    def __init__(self, fs=22050, hop_length=256, fmin=None, n_bins=360, bins_per_octave=60, gamma=0,
                 max_p=1, random=False, update=True, to_db=True, db_to_prob=True, batch_norm=True,
                 var_drop=False):
        """
        Initialize parameters common to all LVQT variants.

        Parameters
        ----------
        fs : int or float
          Number of samples per second of audio
        hop_length : int
          Number of samples between frames
        fmin : float
          Lowest center frequency in basis
        n_bins : int
          Number of basis functions in the filterbank
        bins_per_octave : int
          Number of basis functions per octave
        gamma : float
          Bandwidth offset to smoothly vary Q-factor
        max_p : int
          Kernel size and stride for max pooling operation (1 to disable)
          Note : values which go evenly into the hop length are ideal
               - otherwise, integer divide will result in mini-hops which
               - do not add up exactly to the overall hop length
        random : bool
          Keep the weights random instead of loading in the bases
        update : bool
          Whether to update the weights or ignore the gradients
        to_db : bool
          Convert features from amplitude to decibels
        db_to_prob : bool
          Scale decibel values to be between 0 and 1 if log is taken
        batch_norm : bool
          Perform batch normalization
        var_drop : float
          Doubles as switch (0 to disable) for variational dropout and initial value of log_sigma ^ 2
        """

        # Load PyTorch Module properties
        super(_LVQT, self).__init__()

        # Make parameters accessible
        self.fs = fs
        self.hop_length = hop_length
        self.n_bins = n_bins
        self.bins_per_octave = bins_per_octave
        self.max_p = max_p
        self.random = random
        self.update = update
        self.to_db = to_db
        self.db_to_prob = db_to_prob
        self.batch_norm = batch_norm
        self.var_drop = var_drop

        # Default the minimum frequency
        if fmin is None:
            # Note C1
            fmin = librosa.note_to_hz('C1')
        self.fmin = fmin

        # Default gamma using the procedure defined in
        # librosa.filters.constant_q.vqt documentation
        if gamma is None:
            alpha = 2.0 ** (1.0 / bins_per_octave) - 1
            gamma = 24.7 * alpha / 0.108
        self.gamma = gamma

        # Determine the center frequencies of initialized basis functions
        center_freqs = fmin * (2.0 ** (np.arange(n_bins) / bins_per_octave))
        # Calculate the constant Q factor necessary for the chosen resolution
        Q = 1 / (2.0 ** (1.0 / bins_per_octave) - 1.0)
        # Determine the upper boundary of each filters' bandwidth
        band_bounds = center_freqs * (1 + 0.5 * librosa.filters.window_bandwidth('hann') / Q)
        # Determine the number of invalid filters (past Nyquist)
        num_invalid = int(np.sum(band_bounds > fs / 2.0))

        # Get complex bases and their respective lengths for a variable-Q transform
        basis, lengths = constant_q(sr=fs,
                                    fmin=fmin,
                                    n_bins=n_bins - num_invalid,
                                    bins_per_octave=bins_per_octave,
                                    gamma=gamma,
                                    pad_fft=False,
                                    norm=None)

        # Fill in space leftover from invalid filters with zeros
        zero_filters = np.zeros((num_invalid, basis.shape[-1]))
        zero_lengths = np.zeros(num_invalid)

        # Append any zero filters to the bases
        self.basis = np.concatenate((basis, zero_filters), axis=0)
        self.lengths = np.concatenate((lengths, zero_lengths), axis=0)

        # Stride the amount of samples necessary to take 'max_p' responses per frame
        self.sd1 = self.hop_length // self.max_p
        # Padding to start centered around the first real sample,
        # and end centered around the last real sample
        pd1 = self.basis.shape[1] // 2
        self.pd1 = (pd1, self.basis.shape[1] - pd1)
        # Kernel must be as long as longest basis
        self.ks1 = self.basis.shape[1]

        if self.hop_length % self.max_p != 0:
            # Make sure max pooling divides evenly into hop length
            warnings.warn(f'Max pooling {self.max_p} does not evenly divide into hop length {self.hop_length}.'
                          '\nThis will ruin alignment and may cause unexpected behavior.')

        # Initialize max pooling to take 'max_p' responses per frame and aggregate with max operation
        self.mp = torch.nn.MaxPool2d((1, self.max_p))
        # Initialize batch normalization to normalize the output of each channel
        self.bn = torch.nn.BatchNorm1d(self.n_bins)

    @abstractmethod
    def forward(self, audio):
        """
        Perform the main processing steps for the filterbank.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        feats : Tensor (B x F x T)
          Features calculated for a batch of tracks,
          B - batch size
          F - dimensionality of features (number of bins)
          T - number of time steps (frames)
        """

        # We manually do the padding for the convolutional
        # layer to allow for different front/back padding
        padded_audio = torch.nn.functional.pad(audio, self.pd1)

        # Convolve the audio with the filterbank of real weights
        if self.update:
            # Feed-forward as normal
            feats = self.time_conv(padded_audio)
        else:
            with torch.no_grad():
                # Disable gradient accumulation
                feats = self.time_conv(padded_audio)

        return feats

    @abstractmethod
    def post_proc(self, feats):
        """
        Perform the post-processing steps for the filterbank.

        Parameters
        ----------
        feats : Tensor (B x F x T)
          Features calculated for a batch of tracks,
          B - batch size
          F - dimensionality of features (number of bins)
          T - number of time steps (frames)

        Returns
        ----------
        feats : Tensor (B x F x T)
          Post-processed features for a batch of track.
        """

        # Pad by 1 frame less than the max pooling amount
        pad_amt, half_pad = self.max_p - 1, (self.max_p - 1) // 2
        # Pad the features so any extra frames are not thrown away
        padded_feats = torch.nn.functional.pad(feats, (half_pad, pad_amt - half_pad))
        # Perform max pooling operation
        feats = self.mp(padded_feats)

        if self.to_db:
            # Convert the raw filterbank output to decibels
            feats = torch_amplitude_to_db(feats, to_prob=self.db_to_prob)

        # Number of frames obtained from the filterbank
        num_frames = feats.size(-1)

        if self.batch_norm and not (self.training and num_frames <= 1):
            # TODO - cannot remember why the checked case is necessary
            # Perform batch normalization
            feats = self.bn(feats)

        return feats

    def pad_audio(self, audio):
        """
        Pad audio to squeeze another frame out of trailing samples.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)
        """

        # Determine the remaining number of samples required for an additional frame
        remaining = self.hop_length - ((audio.shape[-1]) % self.hop_length)

        # If there are trailing samples...
        if remaining != self.hop_length:
            # Pad the audio with zeros to fill the remaining samples
            shape = tuple(audio.shape[:-1]) + tuple([remaining])
            filler = torch.zeros(shape).to(audio.device)
            audio = torch.cat((audio, filler), dim=-1)

        return audio

    def get_expected_frames(self, audio):
        """
        Determine the number of frames we expect from provided audio.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        num_frames : int
          Number of frames which will be generated for given audio
        """

        # Number of hops in the audio plus one
        num_frames = audio.shape[-1] // (self.sd1 * self.max_p) + 1

        return num_frames

    @abstractmethod
    def get_real_weights(self):
        """
        Obtain the weights of the real part of the transform.

        Returns
        ----------
        real_weights : Tensor (F x T)
          Weights of the real part of the transform,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        return NotImplementedError

    @abstractmethod
    def get_imag_weights(self):
        """
        Obtain the weights of the imaginary part of the transform.

        Returns
        ----------
        imag_weights : Tensor (F x T)
          Weights of the imaginary part of the transform,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        return NotImplementedError

    def get_mag_weights(self):
        """
        Obtain the magnitude of the complex weights.

        Returns
        ----------
        mag_weights : Tensor (F x T)
          Magnitude of the complex weights,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        real_weights = self.get_real_weights()
        imag_weights = self.get_imag_weights()

        mag_weights = torch.sqrt(real_weights ** 2 + imag_weights ** 2)

        return mag_weights

    def get_comp_weights(self):
        """
        Obtain the complex weights.

        Returns
        ----------
        comp_weights : ndarray (F x T)
          Complex-valued weights as NumPy array,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        real_weights = tensor_to_array(self.get_real_weights())
        imag_weights = tensor_to_array(self.get_imag_weights())

        comp_weights = real_weights + 1j * imag_weights

        return comp_weights

    def visualize(self, save_dir, **kwargs):
        """
        Visualization function which extracts relevant keyword arguments and
        runs all visualization functions.

        Parameters
        ----------
        save_dir : string
          Top-level directory to hold images of all plots
        **kwargs : N/A
          Arguments for generating plots
        """

        # Create directory for time-domain visualization
        time_dir = os.path.join(save_dir, 'time')

        # Extact keyword arguments for time-domain visualization
        time_kwargs = filter_kwargs(['idcs',
                                     'fix_scale',
                                     'include_axis'], **kwargs)

        # Visualize filters in time-domain
        self.visualize_time_domain_complex(time_dir, **time_kwargs)

        # Create directory for frequency-domain visualization
        fft_dir = os.path.join(save_dir, 'fft')

        # Extact keyword arguments for 1D frequency-domain visualization
        fft_1d_kwargs = filter_kwargs(['idcs',
                                       'n_fft',
                                       'include_axis',
                                       'scale_freqs',
                                       'decibels',
                                       'include_negative',
                                       'separate'], **kwargs)

        # Visualize filters in 1D frequency-domain
        self.visualize_freq_domain_fft_1d(fft_dir, **fft_1d_kwargs)

        # Create path for 2D frequency-domain plot
        fft_2d_path = os.path.join(fft_dir, f'all_2d.jpg')

        # Extact keyword arguments for 2D frequency-domain visualization
        fft_2d_kwargs = filter_kwargs(['idcs',
                                       'n_fft',
                                       'sort_by_centroid',
                                       'include_axis',
                                       'scale_freqs',
                                       'include_negative'], **kwargs)

        # Visualize filters in 2D frequency-domain
        self.visualize_freq_domain_fft_2d(fft_2d_path, **fft_2d_kwargs)

    def visualize_time_domain_complex(self, save_dir, idcs=None, fix_scale=False, include_axis=False):
        """
        Plot the time domain filters of the filterbank.

        Parameters
        ----------
        save_dir : string
          Directory under which to save images for each plot
        idcs : list, ndarray or None (optional)
          Specific filter indices to plot rather than plotting all of them
        fix_scale : bool
          Whether to place all filters on the same amplitude scale
        include_axis : bool
          Whether to add X and Y axis and a grid along the Y axis
        """

        # Make sure the provided save directory exists
        os.makedirs(save_dir, exist_ok=True)

        # Obtain the real and imaginary filterbank weights
        real_weights = tensor_to_array(self.get_real_weights())
        imag_weights = tensor_to_array(self.get_imag_weights())

        # Determine the maximum weight magnitude for scaling
        max_weight = max(np.max(np.abs(real_weights)), np.max(np.abs(imag_weights)))

        # Determine the amount of weights in each filter
        num_weights = real_weights.shape[-1]

        # Create ascending indices to loop through
        filter_idcs = np.arange(self.n_bins)

        if idcs is not None:
            # Reduce the indices to those specified by user
            filter_idcs = np.intersect1d(filter_idcs, idcs)

        # Create a figure and axis for plotting
        fig = plt.figure()
        ax = SubplotZero(fig, 111)
        fig.add_subplot(ax)

        if fix_scale:
            # Make the y boundaries of each filter 10% of the maximum weight magnitude
            ax.set_ylim([-1.1 * max_weight, 1.1 * max_weight])

        # Remove right, top, and bottom border
        ax.axis['right'].set_visible(False)
        ax.axis['top'].set_visible(False)
        ax.axis['bottom'].set_visible(False)

        if include_axis:
            # Add X axes at origin
            ax.axis['xzero'].set_visible(True)
            # No X ticks
            ax.set_xticks([])
            # Remove space padding along X axis
            ax.set_xlim([0, num_weights])
            # Add a grid to the axis
            ax.grid(axis='y')
        else:
            # Remove the left border
            ax.axis['left'].set_visible(False)

        # Loop through the filter indices
        for k in filter_idcs:
            if include_axis:
                # Add a title
                ax.set_title(f'Filter Weights ($\mu = {k}$)')

            # Obtain a list of lines on the plot
            lines = ax.get_lines()

            # Check if the real and imaginary lines have already been plotted
            if len(lines):
                # Update the existing real weights
                lines[0].set_ydata(real_weights[k])
                # Update the existing imaginary weights
                lines[1].set_ydata(imag_weights[k])
            else:
                # Make a new line for the real weights
                ax.plot(real_weights[k], color='black', label='Real', alpha=0.75)
                # Make a new line for the imaginary weights
                ax.plot(imag_weights[k], color='purple', label='Imag', alpha=0.75)

            # Construct a path to save an image of the plot
            save_path = os.path.join(save_dir, f'f_{k}.jpg')

            # Save the figure
            fig.savefig(save_path, bbox_inches='tight')

        # Close the figure
        plt.close(fig)

    def visualize_freq_domain_fft_1d(self, save_dir, idcs=None, n_fft=None, include_axis=False,
                                     scale_freqs=False, decibels=False, include_negative=False, separate=True):
        """
        Plot the FFT response of the filterbank and display in 1D fashion.

        Parameters
        ----------
        save_dir : string
          Directory under which to save images for each plot
        idcs : list, ndarray or None (optional)
          Specific filter indices to plot rather than plotting all of them
        n_fft : int or None (optional)
          See np.fft.fft documentation...
        include_axis : bool
          Whether to add X and Y axis and a grid along the Y axis
        scale_freqs : bool
          Whether to leave frequencies as they are or scale to be between [-1, 1]
        decibels : bool
          Whether to convert to dB or leave as amplitude
        include_negative : bool
          Whether to include negative frequencies in the plot
        separate : bool
          Whether to plot the response of each filter separately or all in one plot
        """

        # Make sure the provided save directory exists
        os.makedirs(save_dir, exist_ok=True)

        # Obtain the complex filterbank weights
        comp_weights = self.get_comp_weights()

        # Determine the Nyquist
        nyquist = self.fs / 2

        # Create ascending indices to loop through
        filter_idcs = np.arange(self.n_bins)

        if idcs is not None:
            # Reduce the indices to those specified by user
            filter_idcs = np.intersect1d(filter_idcs, idcs)

        # Scale the X axis to be twice the length of Y axis if negative frequencies are included
        figsize = rcParams['figure.figsize']
        figsize = [(2 ** include_negative) * figsize[0], figsize[1]]

        # Keep track of the default font size
        default_font_size = rcParams['font.size']
        # Change the font size temporarily
        rcParams['font.size'] = 22

        # Create a figure and axis for plotting
        fig = plt.figure(figsize=figsize)
        ax = SubplotZero(fig, 111)
        fig.add_subplot(ax)

        if include_negative:
            # Keep negative frequencies and remove space padding along X axis
            ax.set_xlim(-nyquist ** (not scale_freqs), nyquist ** (not scale_freqs))
        else:
            # Remove negative frequencies and remove space padding along X axis
            ax.set_xlim(0, nyquist ** (not scale_freqs))

        # Remove right and top border
        ax.axis['right'].set_visible(False)
        ax.axis['top'].set_visible(False)

        # Calculate the FFT response for all filters at once
        freqs, resp = fft_response(comp_weights, self.fs, n_fft, decibels)

        # Remove space padding along Y axis
        ax.set_ylim(bottom=-80) if decibels else ax.set_ylim(bottom=0)

        # Make the top boundary of Y axis 10% above maximum
        if decibels:
            ax.set_ylim(top=10 * math.log10(1.1) + np.max(resp))
        else:
            ax.set_ylim(top=1.1 * np.max(resp))

        # Default the X axis label
        xlabel = 'Frequency (Hz)'
        if scale_freqs:
            # Scale the frequencies to be within [-1, 1]
            freqs = freqs / nyquist
            # Update the X axis label
            xlabel = 'Normalized Frequency'

        if include_axis:
            # Add an appropriate label to the Y axis
            ax.set_ylabel('dB') if decibels else ax.set_ylabel('A')
            # Add X axis label
            ax.set_xlabel(xlabel)
            # Add a grid to the axis
            ax.grid(axis='y')
        else:
            # Remove the left and bottom border
            ax.axis['left'].set_visible(False)
            ax.axis['bottom'].set_visible(False)

        # Loop through the filter indices
        for k in filter_idcs:
            if include_axis:
                # Add a title
                ax.set_title(f'FFT Response ($\mu = {k}$)')

            # Obtain a list of lines on the plot
            lines = ax.get_lines()

            # Determine if a new line should be created
            if len(lines) and separate:
                # Update the frequency response
                lines[0].set_ydata(resp[k])
            else:
                # Plot the FFT response
                ax.plot(freqs, resp[k], color='gray', label='FFT Response', alpha=0.75)

            if separate:
                # Construct a path to save an image of the plot
                save_path = os.path.join(save_dir, f'f_{k}.jpg')
                # Save the figure
                fig.savefig(save_path, bbox_inches='tight')

        if not separate:
            # Construct a path to save an image of the plot
            save_path = os.path.join(save_dir, f'all_1d.jpg')
            # Save the figure, now that it is complete
            fig.savefig(save_path)

        # Close the figure
        plt.close(fig)

        # Change the font size back to the default
        rcParams['font.size'] = default_font_size

    def visualize_freq_domain_fft_2d(self, save_path, idcs=None, n_fft=None, sort_by_centroid=False,
                                     include_axis=False, scale_freqs=False, include_negative=False):
        """
        Plot the FFT response (dB) of the filterbank and display in 2D fashion.

        Parameters
        ----------
        save_path : string
          Path to use when saving an image of the plot
        idcs : list, ndarray or None (optional)
          Specific filter indices to plot rather than plotting all of them
        n_fft : int or None (optional)
          See np.fft.fft documentation...
        sort_by_centroid : bool
          Whether to order the filters by ascending spectral centroid;
        include_axis : bool
          Whether to add X and Y axis and a grid along the Y axis
        scale_freqs : bool
          Whether to leave frequencies as they are or scale to be between [-1, 1]
        include_negative : bool
          Whether to include negative frequencies in the plot
        """

        # Make sure the provided save directory exists
        os.makedirs(os.path.dirname(save_path), exist_ok=True)

        # Obtain the complex filterbank weights
        comp_weights = self.get_comp_weights()

        # Calculate the FFT response for all filters at once
        freqs, resp = fft_response(comp_weights, self.fs, n_fft, False)

        # Determine the Nyquist
        nyquist = self.fs / 2

        # Create ascending indices to loop through
        filter_idcs = np.arange(self.n_bins)

        if idcs is not None:
            # Reduce the indices to those specified by user
            filter_idcs = np.intersect1d(filter_idcs, idcs)

        # Keep only filters matching specified indices
        resp = resp[filter_idcs]

        # Determine the number of bins in the response
        num_bins = resp.shape[-1]

        if sort_by_centroid:
            # Get the normalized response in case the filters are not normalized
            norm_resp = resp / (np.expand_dims(np.sum(resp, axis=-1), axis=-1) + EPSILON)

            # Compute the spectral centroid of each filter
            centroids = np.dot(norm_resp[..., num_bins // 2:], freqs[num_bins // 2:])
            centroids = centroids / (np.sum(norm_resp[..., num_bins // 2:], axis=-1) + EPSILON)

            # Sort the filters by spectral centroid of positive frequency response
            resp = resp[np.argsort(centroids)]

        # Convert the frequency response from amplitude to decibels
        resp = librosa.amplitude_to_db(resp, ref=np.max)

        # Line up the response properly for the image
        resp = np.transpose(np.flip(resp, axis=-1))

        # Scale the X axis to be twice the length of Y axis. If negative
        # frequencies are included, the Y axis length will be scaled by two as well
        figsize = rcParams['figure.figsize']
        figsize = [2 * figsize[0], (2 ** include_negative) * figsize[1]]

        # Keep track of the default font size
        default_font_size = rcParams['font.size']
        # Change the font size temporarily
        rcParams['font.size'] = 16

        # Create a figure and axis for plotting
        fig = plt.figure(figsize=figsize)
        ax = SubplotZero(fig, 111)
        fig.add_subplot(ax)

        # Determine the frequency boundaries we are operating with
        y_bounds = [-nyquist ** (not scale_freqs), nyquist ** (not scale_freqs)]

        # Plot the response for all filters as an image
        img = ax.imshow(resp, extent=[0, len(filter_idcs), y_bounds[0], y_bounds[1]], aspect='auto')

        # Remove right and top border
        ax.axis['right'].set_visible(False)
        ax.axis['top'].set_visible(False)

        if include_axis:
            # Add a title
            ax.set_title(f'FFT Response $\\forall$ Filters')
            # No X ticks
            ax.set_xticks([])
            # Set the X axis label
            ax.set_xlabel('Filter Index ($\mu$)')
            # Set the Y axis label
            ax.set_ylabel('Normalized Frequency' if scale_freqs else 'Frequency (Hz)')
            # Add a grid to the axis
            ax.grid(axis='y')
            # Add a colorbar to the figure
            fig.colorbar(img, format='%+2.0f dB')
        else:
            # Remove the left and bottom border
            ax.axis['left'].set_visible(False)
            ax.axis['bottom'].set_visible(False)

        if not include_negative:
            # Trim response for negative frequencies
            ax.set_ylim([0, y_bounds[1]])

        # Save and close the figure
        fig.savefig(save_path, bbox_inches='tight')
        plt.close(fig)

        # Change the font size back to the default
        rcParams['font.size'] = default_font_size

    def sonify(self, save_dir, factor=1, idcs=None):
        """
        Represent each filter as audio and write to disk for listening.

        Parameters
        ----------
        save_dir : string
          Directory under which to save audio for each filter
        factor : int
          Time-stretch factor
        idcs : list, ndarray or None (optional)
          Specific filter indices to sonify rather than sonifying all of them
        """

        # Make sure the provided save directory exists
        os.makedirs(save_dir, exist_ok=True)

        # Obtain the real part of the filterbank weights
        weights = tensor_to_array(self.get_real_weights())
        # Create ascending indices to loop through
        filter_idcs = np.arange(self.n_bins)

        if idcs is not None:
            # Reduce the indices to those specified by user
            filter_idcs = np.intersect1d(filter_idcs, idcs)

        # Initialize an empty array to hold sequence of sonified filters
        sequence = np.array([])

        # Loop through the filter indices
        for k in filter_idcs:
            # Append the new "audio" to the sequence
            sequence = np.append(sequence, weights[k])

            # Construct a path to save the sonified filter
            save_path = os.path.join(save_dir, f'f_{k}.flac')

            # Write the sonified filter
            filter_audio = librosa.effects.time_stretch(weights[k], 1 / factor)
            sf.write(save_path, filter_audio, self.fs, format='flac')

        # Construct a path to save the sequence of sonified filters
        save_path = os.path.join(save_dir, f'sequence.flac')

        # Write the sequence of sonified filters
        sequence = librosa.effects.time_stretch(sequence, 1 / factor)
        sf.write(save_path, sequence, self.fs, format='flac')



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from .utils import *

# Regular imports
from torch.autograd import Variable
from torch.nn.parameter import Parameter

import torch.nn.functional as F
import torch.nn as nn
import torch
import math


class VariationalDropoutConv1d(nn.Conv1d):
    """
    Implements a standard 1D convolutional layer (no bias) with variational dropout,
    as proposed in http://proceedings.mlr.press/v70/molchanov17a.html with heavy
    inspiration from https://github.com/kefirski/variational_dropout.
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride, threshold=math.inf, log_sigma2=-20):
        """
        Initialize the convolutional layer and establish parameter defaults in function signature.

        Parameters
        ----------
        in_channels : int
          Number of feature channels incoming
        out_channels : int
          Number of filters in the convolutional layer
        kernel_size : int
          Number of weights in each filter (receptive field)
        stride : int
          Number of samples between hops
        threshold : float
          Value for log_alpha above which corresponding weights become zero, e.g. 3 -> alpha /approx 20
        log_sigma2 : float
          Initial value of log sigma ^ 2 (determines initial value of alpha), e.g. -10 -> sigma ^ 2 /approx 4.5e-5
          Note : sigma ^ 2 can be thought of as the variance of the weights
               - this can cause problems if set too low
        """

        super(VariationalDropoutConv1d, self).__init__(in_channels, out_channels, kernel_size, stride, bias=False)

        self.threshold = threshold

        # Initialize the variance term with same dimensionality as weights
        self.log_sigma2 = Parameter(torch.FloatTensor(self.weight.size()).fill_(log_sigma2))

        # Initialize the field for the dropout rate
        self.log_alpha = None

    def forward(self, in_feats):
        """
        Feed features through the convolutional layer, adding Gaussian noise
        (with learned individual rates) to each weight during training. During
        inference, weights with corresponding log alpha (dropout rate) beyond
        the chosen threshold are zeroed.

        Parameters
        ----------
        in_feats : Tensor (B x C x T)
          Incoming features for a batch,
          B - batch size
          C - number of input channels
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        out_feats : Tensor (B x F x N)
          Outgoing features calculated for a batch,
          B - batch size
          F - dimensionality of features (number of filters)
          N - number of time/spatial steps (frames)
        """

        # Infer log alpha from log sigma ^ 2 and the current weights
        # sigma ^ 2 = alpha * theta ^ 2 -> log alpha = log sigma ^ 2 - log theta ^ 2
        self.log_alpha = self.log_sigma2 - torch.log(self.weight ** 2 + EPSILON)

        if self.training:
            # Convolve the input features with the mean of the weights to obtain the mean of the output
            gamma = F.conv1d(in_feats, weight=self.weight, stride=self.stride)

            # Compute the variance of the weights
            sigma2 = torch.exp(self.log_sigma2)

            # Convolve the (squared) input features with the variance to obtain the variance of the output
            delta = F.conv1d(in_feats ** 2, weight=sigma2, stride=self.stride)

            # Compute the standard deviation of the output
            sqrt_delta = torch.sqrt(delta + EPSILON)

            # Sample from standard normal distribution
            noise = torch.cuda.FloatTensor(*gamma.size(), device=in_feats.device).normal_()

            # Add the Gaussian noise to the outgoing features
            out_feats = gamma + sqrt_delta * noise
        else:
            # Determine which weights should be zeroed
            mask = self.log_alpha > self.threshold
            # Feed the input features through the convolutional layer with masked weights
            out_feats = F.conv1d(in_feats,
                                 weight=self.weight.masked_fill(mask, 0),
                                 stride=self.stride)

        return out_feats

    def kld(self):
        """
        Compute the approximate KL-divergence of the current weights.

        Returns
        ----------
        kld : float
          KL-divergence averaged across all weights of the layer
        """

        # Terms for approximate KL-divergence
        k = [0.63576, 1.87320, 1.48695]

        # Compute the approximate negative KL-divergence
        nkl = (k[0] * torch.sigmoid(k[1] + k[2] * self.log_alpha) -
               0.5 * torch.log(1 + torch.exp(-self.log_alpha)) - k[0])

        # Flip sign and average the KL-divergence across the weights
        kld = -torch.sum(nkl) / (self.in_channels * self.out_channels * self.kernel_size[0])

        return kld



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from .lhvqt import *

# Regular imports
import torch.nn as nn
import torch


class LHVQT_COMB(LHVQT):
    """
    Harmonic learnable filterbank where the filters
    associated with each harmonic are collapsed into one.
    """

    def __init__(self, fmin=None, harmonics=None, lvqt=None, **kwargs):
        """
        Initialize parameters necessary to construct a harmonic learnable filterbank.

        Parameters
        ----------
        fmin : float
          Lowest center frequency in basis
        harmonics : list of int
          Harmonics to compute
        lvqt : type
          Class definition of chosen lower-level LVQT module
        **kwargs : N/A
          Any parameters intended for the lower-level LVQT module
        """

        # Load PyTorch Module properties
        super(LHVQT_COMB, self).__init__(fmin=fmin, harmonics=harmonics, lvqt=lvqt, **kwargs)

        # Obtain a pointer to the lower-level modules
        lvqt_modules = self.get_modules()

        # Reset the collection of filterbanks
        self.tfs = nn.Module()
        # Use the filterbank associated with the lowest harmonic to match hyperparameters
        self.tfs.add_module('lvqt', lvqt_modules[0])

        # Get the weights from the lowest harmonic
        harmonic_weights = lvqt_modules[0].time_conv.weight

        # Loop through harmonics
        for h in range(1, len(self.harmonics)):
            # Get the filterbank weights associated with the harmonic
            weights = lvqt_modules[h].time_conv.weight

            # Pad the weights to match the size of the lowest harmonic
            pad_amt = harmonic_weights.size(-1) - weights.size(-1)
            # Distribute the padding among both sides of the weights
            pad = (pad_amt // 2, pad_amt - pad_amt // 2)
            # Pad and add the weights to the summed harmonic weights
            harmonic_weights = harmonic_weights + nn.functional.pad(weights, pad=pad)

        # Replace the filterbank weights with the summer harmonic weights
        self.tfs.lvqt.time_conv.weight = torch.nn.Parameter(harmonic_weights)

    def forward(self, wav):
        """
        Perform the main processing steps for the harmonic filterbank.

        Parameters
        ----------
        wav : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        feats : Tensor (B x H x F x T)
          Input features for a batch of tracks,
          B - batch size
          H - number of harmonics (a.k.a. channels)
          F - dimensionality of features
          T - number of time steps (frames)
        """

        # Run the audio through the singular module
        feats = self.tfs.lvqt(wav).unsqueeze(0)
        # Switch harmonic and batch dimension
        feats = feats.transpose(1, 0)

        return feats

    def get_expected_frames(self, audio):
        """
        Determine the number of frames we expect from provided audio.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        num_frames : int
          Number of frames which will be generated for given audio
        """

        # Number of hops in the audio plus one
        num_frames = self.tfs.lvqt.get_expected_frames(audio)

        return num_frames

    def visualize(self, save_dir, **kwargs):
        """
        Perform visualization steps.

        Parameters
        ----------
        save_dir : string
          Top-level directory to hold images of all plots
        **kwargs : N/A
          Arguments for generating plots
        """

        # Visualize the singular module
        self.tfs.lvqt.visualize(save_dir, **kwargs)

    def sonify(self, save_dir, **kwargs):
        """
        Perform sonification steps.

        Parameters
        ----------
        save_dir : string
          Top-level directory to hold images of all plots
        **kwargs : N/A
          Arguments for generating plots
        """

        # Sonify the singular module
        self.tfs.lvqt.sonify(save_dir, **kwargs)



from setuptools import setup, find_packages

# TODO - print(lhvqt.__version__) - what else am I missing?
# TODO - none-any.whl?

setup(
    name='lhvqt',
    url='https://github.com/cwitkowitz/LHVQT',
    author='Frank Cwitkowitz',
    author_email='fcwitkow@ur.rochester.edu',
    packages=find_packages(),
    python_requires='>=3.7',
    install_requires=['numpy', 'librosa', 'torch', 'matplotlib', 'soundfile'],
    version='0.5.0',
    license='MIT',
    description='Frontend filterbank learning module with HVQT initialization capabilities',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown'
)



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from .lvqt_orig import *

# Regular imports
import librosa
import torch
import os


class LHVQT(torch.nn.Module):
    """
    Learnable filterbank with various HVQT initialization capabilties.
    """

    def __init__(self, fmin=None, harmonics=None, lvqt=None, **kwargs):
        """
        Initialize parameters necessary to construct a harmonic learnable filterbank.

        Parameters
        ----------
        fmin : float
          Lowest center frequency in basis
        harmonics : list of ints
          Specific harmonics to stack across the harmonic dimension
        lvqt : type
          Class definition of chosen lower-level LVQT module
        **kwargs : N/A
          Any parameters intended for the lower-level LVQT module
        """

        # Load PyTorch Module properties
        super(LHVQT, self).__init__()

        # Default the minimum frequency
        if fmin is None:
            # Note C1
            fmin = librosa.note_to_hz('C1')
        self.fmin = fmin

        # Default the harmonics to those used in the DeepSalience paper
        if harmonics is None:
            harmonics = [0.5, 1, 2, 3, 4, 5]
        harmonics.sort()
        self.harmonics = harmonics

        # Default the class definition for the lower-level module
        if lvqt is None:
            # Classic LVQT module
            lvqt = LVQT

        # Create a PyTorch Module to hold LVQTs
        self.tfs = torch.nn.Module()

        # Loop through harmonics
        for h in range(len(self.harmonics)):
            # Name module according to index of harmonic
            mod_name = 'lvqt%d' % (h + 1)
            # Calculate the lowest center frequency of this LVQT
            fmin = float(harmonics[h]) * self.fmin
            # Create and add the LVQT to the LHVQT module
            self.tfs.add_module(mod_name, lvqt(fmin=fmin, **kwargs))

    def get_modules(self):
        """
        Helper function to extract pointers to the lower-level modules.

        Returns
        ----------
        lvqt_modules : list of _LVQT
          Pointers to each module in the harmonic filterbank
        """

        # Obtain a pointer to the lower-level modules
        lvqt_modules = torch.nn.Sequential(*list(self.tfs.children()))

        return lvqt_modules

    def forward(self, wav):
        """
        Perform the main processing steps for the harmonic filterbank.

        Parameters
        ----------
        wav : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        feats : Tensor (B x H x F x T)
          Input features for a batch of tracks,
          B - batch size
          H - number of harmonics (a.k.a. channels)
          F - dimensionality of features
          T - number of time steps (frames)
        """

        # Initialize a list to hold each harmonic transform
        tf_list = []

        # Obtain a pointer to the lower-level modules
        lvqt_modules = self.get_modules()

        # Loop through harmonics
        for h in range(len(self.harmonics)):
            # Take the transform at each harmonic
            tf = lvqt_modules[h](wav)
            # Add a harmonic dimension
            tf = tf.unsqueeze(0)
            # Add the transform to the list
            tf_list.append(tf)

        # Combine the transforms together along harmonic dimension
        feats = torch.cat(tf_list, dim=0)
        # Switch harmonic and batch dimension
        feats = feats.transpose(1, 0)

        return feats

    def get_expected_frames(self, audio):
        """
        Determine the number of frames we expect from provided audio.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        num_frames : int
          Number of frames which will be generated for given audio
        """

        # Obtain a pointer to the lower-level modules
        lvqt_modules = self.get_modules()

        # Gather expected counts from lower-level modules
        num_frames = [lvqt_modules[h].get_expected_frames(audio)
                      for h in range(len(self.harmonics))]

        # Check if count of each module is the same
        if len(set(num_frames)) == 1:
            # Collapse the expected frame counts
            num_frames = num_frames[0]

        return num_frames

    def visualize(self, save_dir, **kwargs):
        """
        Perform visualization steps for each lower-level module.

        Parameters
        ----------
        save_dir : string
          Top-level directory to hold images of all plots
        **kwargs : N/A
          Arguments for generating plots
        """

        # Obtain a pointer to the lower-level modules
        lvqt_modules = self.get_modules()

        # Loop through harmonics
        for h in range(len(self.harmonics)):
            # Construct a path to the directory for the harmonic
            h_dir = os.path.join(save_dir, f'h_{self.harmonics[h]}') if len(self.harmonics) > 1 else save_dir
            # Visualize the harmonic
            lvqt_modules[h].visualize(h_dir, **kwargs)

    def sonify(self, save_dir, **kwargs):
        """
        Perform sonification steps for each lower-level module.

        Parameters
        ----------
        save_dir : string
          Top-level directory to hold images of all plots
        **kwargs : N/A
          Arguments for generating audio
        """

        # Obtain a pointer to the lower-level modules
        lvqt_modules = self.get_modules()

        # Loop through harmonics
        for h in range(len(self.harmonics)):
            # Construct a path to the directory for the harmonic
            h_dir = os.path.join(save_dir, f'h_{self.harmonics[h]}') if len(self.harmonics) > 1 else save_dir
            # Sonify the harmonic
            lvqt_modules[h].sonify(h_dir, **kwargs)



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from .lvqt import _LVQT
from .variational import *

# Regular imports
import numpy as np
import torch


class LVQT(_LVQT):
    """
    Implements an extension of the original LVQT module. In this version, only the real channel is used.
    """

    def __init__(self, **kwargs):
        """
        Initialize LVQT parameters and the PyTorch processing modules.

        Parameters
        ----------
        See _LVQT class...
        """

        super(LVQT, self).__init__(**kwargs)

        # One channel (audio) coming in
        nf_in = 1
        # One channel (real) for each bin going out
        nf_out = self.n_bins

        # Initialize the 1D convolutional filterbank
        if self.var_drop:
            self.time_conv = VariationalDropoutConv1d(in_channels=nf_in, out_channels=nf_out,
                                                      kernel_size=self.ks1, stride=self.sd1,
                                                      log_sigma2=self.var_drop)
        else:
            self.time_conv = torch.nn.Conv1d(in_channels=nf_in, out_channels=nf_out,
                                             kernel_size=self.ks1, stride=self.sd1,
                                             bias=False)

        if not self.random:
            # Get the real weights from the complex valued bases
            real_weights = np.real(self.basis)
            # Add a channel dimension to the weights
            real_weights = torch.Tensor(real_weights).view(nf_out, 1, self.ks1)
            # Manually set the Conv1d parameters to the weights
            self.time_conv.weight = torch.nn.Parameter(real_weights)

    def forward(self, audio):
        """
        Perform the main processing steps for the filterbank.

        Parameters
        ----------
        audio : Tensor (B x 1 x T)
          Audio for a batch of tracks,
          B - batch size
          T - number of samples (a.k.a. sequence length)

        Returns
        ----------
        feats : Tensor (B x F x T)
          Features calculated for a batch of tracks,
          B - batch size
          F - dimensionality of features (number of bins)
          T - number of time steps (frames)
        """

        # Run the standard convolution steps
        feats = super().forward(audio)

        # Perform post-processing steps
        feats = self.post_proc(feats)

        return feats

    def get_real_weights(self):
        """
        Obtain the real-valued weights.

        Returns
        ----------
        real_weights : Tensor (F x T)
          Weights of the real part of the transform,
          F - number of frequency bins
          T - number of time steps (samples)
        """

        real_weights = self.time_conv.weight
        real_weights = real_weights.squeeze()

        return real_weights

    def get_imag_weights(self):
        """
        Return zeros matching size of real weights

        Returns
        ----------
        imag_weights : Tensor (F x T)
          Weights of the imaginary part of the transform (all zeros here),
          F - number of frequency bins
          T - number of time steps (samples)
        """

        real_weights = self.get_real_weights()
        imag_weights = torch.zeros(real_weights.size())
        imag_weights = imag_weights.to(real_weights.device)

        return imag_weights
